#
# Aura Engine API (https://gitlab.servus.at/aura/engine-api)
#
# Copyright (C) 2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import threading
import requests

from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy     import create_engine

from models import PlayLog, ActivityLog




class SyncJob(threading.Thread):
    """
    Job periodically checking for data at one of the main nodes to be synchronized.

    This job should be running on a "Sync Node" only.
    """

    config = None
    logger = None
    app = None
    exit_event = None
    sync_interval = None
    sync_batch_size = None

    Session = None


    def __init__(self, config, logger, app):
        """
        Initialize Job.
        """
        self.config = config
        self.logger = logger
        self.app = app

        threading.Thread.__init__(self)
        self.exit_event = threading.Event()
        self.sync_interval = self.config.get("sync_interval")
        self.sync_batch_size = self.config.get("sync_batch_size")
        self.logger.info("Initialized Sync Job - Synchronizing API Nodes every %s seconds and with a max batch-size of %s." % (self.sync_interval, self.sync_batch_size))

        # Create a scoped local database session to be thread safe
        engine = create_engine(self.config.get_database_uri())
        session_factory = sessionmaker(autoflush=True, autocommit=False, bind=engine)
        self.Session = scoped_session(session_factory)
    

    def run(self):
        """
        Starts the Job.
        """
        self.synchronize()
        while not self.exit_event.wait(self.sync_interval):
            self.synchronize()
            self.logger.info("Sync cycle done.")


    def synchronize(self):
        """
        Performs the synchronization with the main nodes.
        """
        entries = None
        synced_entries = 0
        unsynced_sources = self.get_unsynced_history()
        self.logger.info("\nSynchronization API Nodes: There are %s sources open to be synced." % len(unsynced_sources))

        for i in range(len(unsynced_sources)):

            source = unsynced_sources[i]
            self.logger.info("Syncing source %s which is unsynced since %s" % (source.source_number, source.log_time))

            # Store the next source to build a datetime range
            next_source = None
            if i+1 < len(unsynced_sources):
                next_source = unsynced_sources[i+1]

            try:
                # Get unsynced entries
                entries = self.get_unsynced_entries(source, next_source)

                while len(entries) > 0:
                    if not entries: self.logger.info("Retrieved no entries to be synced")
                    else: self.logger.info("Retrieved %s playlogs to be synced" % len(entries))

                    # Store unsynced entries locally
                    for entry in entries:

                        try:
                            playlog = PlayLog(entry)
                            self.Session.add(playlog)
                            self.Session.commit()
                            self.logger.info("Stored synced playlog for '%s'" % playlog.track_start)
                            synced_entries += 1
                        except sqlalchemy.exc.IntegrityError as e:
                            self.logger.info("Playlog for '%s' is already existing in local database. Skipping..." % playlog.track_start)
                
                        # Sleep a little to keep the effective load down low
                        time.sleep(self.config.get("sync_step_sleep"))

                    # Get some more unsynced entries, if available
                    entries = self.get_unsynced_entries(source, next_source)

                # Update state of the source activity log to `synced=true`
                try:
                    source.is_synced = True
                    self.Session.add(source)
                    self.Session.commit()
                    self.logger.info("Sync for source %s finalized!" % source.source_number)
                except Exception as e:
                    self.logger.error("Cannot finalize sync state for source=%s - Reason: %s" % (source.source_number, str(e)))

                # For now, let it be ... if there's more to sync let's do it in the next cycle
                self.logger.info("... sucessfully synchronized %s playlogs!\n\n" % synced_entries)

            except Exception:
                self.logger.info("Error while syncing entries. Maybe there's a connectivity issue. Aborting cycle ...")
                return


    def get_unsynced_history(self):
        """
        Retrieves all sources with un-synced states
        """      
        self.Session.commit()
        unsynced = self.Session.query(ActivityLog).filter(ActivityLog.is_synced == False).\
            order_by(ActivityLog.log_time.asc())
        return unsynced.all()


    def get_unsynced_entries(self, source, next_source):
        """
        Retrieve unsynced entries from main node
        """
        response = None
        url = None
        to_time = None
        if next_source:
            to_time = next_source.log_time

        try:
            params = { "page": "0", "limit": self.sync_batch_size, "skip_synced": "true", "from_date": source.log_time, "to_date": to_time}
            url = self.get_url(source.source_number)
            response = requests.get(url, params=params)
            if response.status_code == 200:
                self.logger.info("Successfully retrieved unsynced entries from '%s'" % url)
                return response.json()
            else:
                self.logger.error("Invalid status code while getting unsynced entries from remote API: " + str(response.status_code))
        except Exception as e:
            self.logger.error("Error while getting unsynced entries from remote API '%s'!\n%s" % (url, str(e)))
            raise e


    def get_url(self, source_number):
        """
        Builds an URL for the remote API.
        """
        url = self.config.get("main_host_" + str(source_number))
        url += self.config.get("sync_api_get_playlog")
        return url


    def exit(self):
        """
        Called when the application shuts down.
        """
        self.exit_event.set()




